@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2013-2017 Canonical, Ltd.
+ * Copyright (C) 2013-2021 Canonical, Ltd.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -24,12 +24,19 @@
  */
 #include "stress-ng.h"
 
+static const stress_help_t help[] = {
+	{ NULL,	"chown N",	"start N workers thrashing chown file ownership" },
+	{ NULL,	"chown-ops N",	"stop chown workers after N bogo operations" },
+	{ NULL,	NULL,		NULL }
+};
+
 /*
  *  do_fchown()
  *	set ownership different ways
  */
 static int do_fchown(
 	const int fd,
+	const int bad_fd,
 	const bool cap_chown,
 	const uid_t uid,
 	const gid_t gid)
@@ -38,28 +45,34 @@ static int do_fchown(
 
 	if (fchown(fd, uid, gid) < 0)
 		return -errno;
-	if (fchown(fd, -1, gid) < 0)
+	if (fchown(fd, (uid_t)-1, gid) < 0)
 		return -errno;
-	if (fchown(fd, uid, -1) < 0)
+	if (fchown(fd, uid, (gid_t)-1) < 0)
 		return -errno;
-	if (fchown(fd, -1, -1) < 0)
+	if (fchown(fd, (uid_t)-1, (gid_t)-1) < 0)
 		return -errno;
 
 	if (cap_chown)
 		return 0;
-	if (fchown(fd, 0, 0) == 0)
+	if (fchown(fd, (uid_t)0, (gid_t)0) == 0)
 		goto restore;
 	if (errno != EPERM)
 		goto restore;
-	if (fchown(fd, -1, 0) == 0)
+	if ((fchown(fd, (uid_t)-1, (gid_t)0) == 0) && (errno != EPERM))
 		goto restore;
 	if (errno != EPERM)
 		goto restore;
-	if (fchown(fd, 0, -1) == 0)
+	if (fchown(fd, (uid_t)0, (gid_t)-1) == 0)
 		goto restore;
 	if (errno != EPERM)
 		goto restore;
 
+	/*
+	 *  Exercise fchown with invalid fd
+	 */
+	ret = fchown(bad_fd, uid, gid);
+	(void)ret;
+
 	return 0;
 
 restore:
@@ -75,6 +88,7 @@ restore:
  *	set ownership different ways
  */
 static int do_chown(
+	int (*chown_func)(const char *pathname, uid_t owner, gid_t group),
 	const char *filename,
 	const bool cap_chown,
 	const uid_t uid,
@@ -82,26 +96,26 @@ static int do_chown(
 {
 	int ret, tmp;
 
-	if (chown(filename, uid, gid) < 0)
+	if (chown_func(filename, uid, gid) < 0)
 		return -errno;
-	if (chown(filename, -1, gid) < 0)
+	if (chown_func(filename, (uid_t)-1, gid) < 0)
 		return -errno;
-	if (chown(filename, uid, -1) < 0)
+	if (chown_func(filename, uid, (gid_t)-1) < 0)
 		return -errno;
-	if (chown(filename, -1, -1) < 0)
+	if (chown_func(filename, (uid_t)-1, (gid_t)-1) < 0)
 		return -errno;
 
 	if (cap_chown)
 		return 0;
-	if (chown(filename, 0, 0) == 0)
+	if (chown_func(filename, (uid_t)0, (gid_t)0) == 0)
 		goto restore;
 	if (errno != EPERM)
 		goto restore;
-	if (chown(filename, -1, 0) == 0)
+	if ((chown_func(filename, (uid_t)-1, (gid_t)0) == 0) && (errno != EPERM))
 		goto restore;
 	if (errno != EPERM)
 		goto restore;
-	if (chown(filename, 0, -1) == 0)
+	if (chown_func(filename, (uid_t)0, (gid_t)-1) == 0)
 		goto restore;
 	if (errno != EPERM)
 		goto restore;
@@ -110,7 +124,7 @@ static int do_chown(
 
 restore:
 	tmp = errno;
-	ret = chown(filename, uid, gid);
+	ret = chown_func(filename, uid, gid);
 	(void)ret;
 
 	return -tmp;
@@ -120,11 +134,12 @@ restore:
  *  stress_chown
  *	stress chown
  */
-int stress_chown(const args_t *args)
+static int stress_chown(const stress_args_t *args)
 {
 	const pid_t ppid = getppid();
 	int fd = -1, rc = EXIT_FAILURE, retries = 0;
-	char filename[PATH_MAX], dirname[PATH_MAX];
+	const int bad_fd = stress_get_bad_fd();
+	char filename[PATH_MAX], pathname[PATH_MAX];
 	const uid_t uid = getuid();
 	const gid_t gid = getgid();
 	bool cap_chown = false;
@@ -135,11 +150,12 @@ int stress_chown(const args_t *args)
 	/*
 	 *  Allow for multiple workers to chown the *same* file
 	 */
-	stress_temp_dir(dirname, sizeof(dirname), args->name, ppid, 0);
-	if (mkdir(dirname, S_IRUSR | S_IRWXU) < 0) {
+	stress_temp_dir(pathname, sizeof(pathname), args->name, ppid, 0);
+	if (mkdir(pathname, S_IRUSR | S_IRWXU) < 0) {
 		if (errno != EEXIST) {
 			rc = exit_status(errno);
-			pr_fail_err("mkdir");
+			pr_fail("%s: mkdir %s failed, errno=%d (%s)\n",
+				args->name, pathname, errno, strerror(errno));
 			return rc;
 		}
 	}
@@ -149,7 +165,8 @@ int stress_chown(const args_t *args)
 	if (args->instance == 0) {
 		if ((fd = creat(filename, S_IRUSR | S_IWUSR)) < 0) {
 			rc = exit_status(errno);
-			pr_fail_err("creat");
+			pr_fail("%s: creat %s failed, errno=%d (%s)\n",
+				args->name, filename, errno, strerror(errno));
 			goto tidy;
 		}
 	} else {
@@ -158,7 +175,18 @@ int stress_chown(const args_t *args)
 			if ((fd = open(filename, O_RDWR, S_IRUSR | S_IWUSR)) > - 1)
 				break;
 
+#if defined(__NetBSD__)
+			/* For some reason usleep blocks */
+			(void)shim_sched_yield();
+			retries = 0;
+#else
 			(void)shim_usleep(100000);
+#endif
+			/* Timed out, then give up */
+			if (!keep_stressing_flag()) {
+				rc = EXIT_SUCCESS;
+				goto tidy;
+			}
 			if (++retries >= 100) {
 				pr_err("%s: chown: file %s took %d "
 					"retries to open and gave up "
@@ -166,24 +194,22 @@ int stress_chown(const args_t *args)
 					args->name, filename, retries, args->instance);
 				goto tidy;
 			}
-			/* Timed out, then give up */
-			if (!g_keep_stressing_flag) {
-				rc = EXIT_SUCCESS;
-				goto tidy;
-			}
 		}
 	}
 
+	stress_set_proc_state(args->name, STRESS_STATE_RUN);
+
 	do {
 		int ret;
 
-		ret = do_fchown(fd, cap_chown, uid, gid);
+		ret = do_fchown(fd, bad_fd, cap_chown, uid, gid);
+		if ((ret < 0) && (ret != -EPERM))
+			pr_fail("%s: fchown failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
+
+		ret = do_chown(chown, filename, cap_chown, uid, gid);
 		if (ret < 0) {
-			pr_fail_err("fchown");
-		}
-		ret = do_chown(filename, cap_chown, uid, gid);
-		if (ret < 0) {
-			if (ret == -ENOENT || errno == -ENOTDIR) {
+			if ((ret == -ENOENT) || (ret == -ENOTDIR)) {
 				/*
 				 * File was removed during test by
 				 * another worker
@@ -191,17 +217,41 @@ int stress_chown(const args_t *args)
 				rc = EXIT_SUCCESS;
 				goto tidy;
 			}
-			pr_fail_err("chown");
+			if (ret != -EPERM)
+				pr_fail("%s: chown %s failed, errno=%d (%s)\n",
+					args->name, filename, errno, strerror(errno));
+		}
+		ret = do_chown(lchown, filename, cap_chown, uid, gid);
+		if (ret < 0) {
+			if ((ret == -ENOENT) || (ret == -ENOTDIR)) {
+				/*
+				 * File was removed during test by
+				 * another worker
+				 */
+				rc = EXIT_SUCCESS;
+				goto tidy;
+			}
+			if (ret != -EPERM)
+				pr_fail("%s: chown %s failed, errno=%d (%s)\n",
+					args->name, filename, errno, strerror(errno));
 		}
 		inc_counter(args);
-	} while (keep_stressing());
+	} while (keep_stressing(args));
 
 	rc = EXIT_SUCCESS;
 tidy:
+	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
+
 	if (fd >= 0)
 		(void)close(fd);
 	(void)unlink(filename);
-	(void)rmdir(dirname);
+	(void)rmdir(pathname);
 
 	return rc;
 }
+
+stressor_info_t stress_chown_info = {
+	.stressor = stress_chown,
+	.class = CLASS_FILESYSTEM | CLASS_OS,
+	.help = help
+};
